#!/usr/bin/env ruby
# Copyright (C) 2009-2010  Adam Sloboda
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

##
# Attention: now listens on AF_UNIX socket /tmp/osd.socket
# instead of netcat you can use "socat - unix-connect:/tmp/osd.socket"

## CONF
path = '/home/as/bin/img/osd/'
BRIGHTNESS_OSD_ICON = path+'notification-display-brightness-high.svg'
VOLUME_OSD_ICON = path+'notification-audio-volume-high.svg'
VOLUME_ICONS = [ path+'notification-audio-volume-off.svg',
                 path+'notification-audio-volume-low.svg',
                 path+'notification-audio-volume-medium.svg',
                 path+'notification-audio-volume-high.svg' ]
VOLUME_MUTE = path+'notification-audio-volume-muted.svg'

## TODO: PROTOTYPE of independent "workspace" switching for 2 displays
# echo wm switch <0-1> <1-4> | nc -U /tmp/osd.socket
# bind to ctrl+f1..4 and win+1..4 (xbindkeys)
# there are more commands:
# - swap between displays,
# - swap full workspace,
# - swap any combination of display/workspace
# - "screens" command to update status
# uses 5 native for 4 independent workspaces -- workspace 0 is always active!

# TODO better command names
# TODO if this crashes, @loaded have to be set to current values
# i don't want to touch disk
# TODO netwmpager has trouble with updates
class Screen
  attr_reader :x, :y, :w

  def initialize(w,h,x,y)
    @w,@h,@x,@y = w.to_i, h.to_i, x.to_i, y.to_i
  end
end

class Manager
  def initialize
    # windows will be moved to workspace 1 when switching
    @loaded = [1, 1]
    @history = [nil, nil]
    screens
  end

  def update
    workspaces
    windows
  end

  def show(args)
    # update screens - this is quite slow
    if args[0] == 'screens'
      screens
    else
      # update window/workspace data
      update

      # TODO
      #require 'time'
      #t = Time.new
      #puts t.sec, t.usec

      if args[0] == 'switch'
        # switch to workspace args[2] on display args[1]
        if args.length == 3
          switch(args[1].to_i, args[2].to_i)
        end
      elsif args[0] == 'swap'
        # swap current halves
        swap
      elsif args[0] == 'swap_displays'
        # swap any combination of halves
        if args.length == 5
          swap_displays(args[1].to_i, args[2].to_i, args[3].to_i, args[4].to_i)
        end
      elsif args[0] == 'swap_workspace'
        # swap full workspaces
        if args.length == 3
          swap_workspace(args[1].to_i, args[2].to_i)
        end
      elsif args[0] == 'set_loaded'
        if args.length == 3
          set_loaded(args[1].to_i, args[2].to_i)
        end
      end

      #t = Time.new
      #puts t.sec, t.usec
    end
  end

  def set_loaded(a, b)
    @loaded = [a, b]
  end

  # implemented for horizontal non-overlapping views
  # returns corresponding display for window
  def get_screen(x)
    @screens.each_index do |i|
      # TODO maybe also 5-10 pixels less than scr.w?
      return i if x - @screens[i].x < @screens[i].w
    end
  end

  def windows
    # load window lists
    `wmctrl -lG`.each do |line|
      cols = line.split
      d = cols[1].to_i
      if d != -1
        @workspaces[d] << [cols[0], # ID
                           cols[2].to_i, cols[3].to_i, # x, y
                           cols[4].to_i, cols[5].to_i, # w, h
                           get_screen(cols[2].to_i)]
      end
    end
  end

  def workspaces
    workspaces = `wmctrl -d`.split("\n")
    @workspaces = Array.new(workspaces.length) { Array.new }
    @current = workspaces.index { |line| /^\d+ +\*/.match(line) }
  end

  def screens
    @screens = []
    screen = /(\d+)x(\d+)\+(\d+)\+(\d+)/
    `xrandr -q`.each do |line|
      result = screen.match(line)
      if result
        @screens << Screen.new(*result[1..-1])
      end
    end
  end

  # swap whole workspaces
  def swap_workspace(a, b)
    @workspaces[a].each { |win| `wmctrl -i -r #{win[0]} -t #{b}` }
    @workspaces[b].each { |win| `wmctrl -i -r #{win[0]} -t #{a}` }
  end

  # swap windows on current displays - implemented for 2 horizontal displays
  def swap
    return if @screens.length != 2
    return if @screens[0].x == @screens[1].x

    @workspaces[@current].each do |win|
      i = win[5]
      relx = win[1] - @screens[i].x
      rely = win[2] - @screens[i].y
      x = @screens[(i+1) % 2].x + relx - 10 # TODO why are reported values +10?
      y = @screens[(i+1) % 2].y + rely - 46 # and +46?
      `wmctrl -i -r #{win[0]} -e 0,#{x},#{y},-1,-1`
    end
  end

  # swap display with another workspace - implemented for 2 horizontal displays
  def switch(i, target)
    return if i>0 and @screens.length != 2
    return if i>0 and @screens[0].x == @screens[1].x

    if target == @loaded[i]
      if @history[i] # switch back
        target = @history[i]
        @history[i] = @loaded[i]
      else # ignore, we're already there
        return
      end
    else
      @history[i] = @loaded[i]
    end

    @workspaces[0].each do |win|
      if win[5] == i
        `wmctrl -i -r #{win[0]} -t #{@loaded[i]}`
      end
    end

    @workspaces[target].each do |win|
      if win[5] == i
        `wmctrl -i -r #{win[0]} -t 0`
      end
    end

    @loaded[i] = target
  end

  def swap_displays(d1, w1, d2, w2)
    return if d1==d2 and w1==w2

    return if @screens.length != 2
    return if @screens[0].x == @screens[1].x

    @workspaces[w1].each do |win|
      if win[5] == d1
        `wmctrl -i -r #{win[0]} -t #{w2}`

        # move window, we changed displays
        if d1 != d2
          relx = win[1] - @screens[d1].x
          rely = win[2] - @screens[d1].y
          # TODO why are reported values +10 and +46?
          x = @screens[d2].x + relx - 10
          y = @screens[d2].y + rely - 46

          `wmctrl -i -r #{win[0]} -e 0,#{x},#{y},-1,-1`
        end
      end
    end

    @workspaces[w2].each do |win|
      if win[5] == d2
        `wmctrl -i -r #{win[0]} -t #{w1}`

        # move window, we changed displays
        if d1 != d2
          relx = win[1] - @screens[d2].x
          rely = win[2] - @screens[d2].y
          # TODO why are reported values +10 and +46?
          x = @screens[d1].x + relx - 10
          y = @screens[d1].y + rely - 46

          `wmctrl -i -r #{win[0]} -e 0,#{x},#{y},-1,-1`
        end
      end
    end
  end

end

## CLASS
require 'gtk2'
require 'socket'

require 'filesystem'

$activate = Proc.new { |item|
  $fstab.each do |m|
    if m[FsTab::MENU] == item
      cmd = "umount "
      if m[FsTab::TYPE][0,4] == 'fuse'
        cmd = "fusermount -u "
      end

      if system(cmd+m[FsTab::MOUNTPOINT])
        `notify-send -i /usr/share/icons/gnome/48x48/devices/gtk-harddisk.png "#{m[FsTab::SPEC]}" "successfully unmounted \"#{m[FsTab::MOUNTPOINT]}\""`
      else
        `notify-send -i error "#{m[FsTab::SPEC]}" "can't umount \"#{m[FsTab::MOUNTPOINT]}\""`
      end
      break
    end
  end
}

class FsTab < Array
  SPEC = 0
  MOUNTPOINT = 1
  TYPE = 2
  OPT = 3
  DUMP = 4
  PASS = 5
  MENU = 6

  def initialize
    File.open('/proc/mounts').each_line do |l|
      entry = l.split
      self << entry if valid_dev(entry)
    end
  end

  def match_dev(dev, string)
    dev[0, string.length] == string
  end

  # some validation from xfce4-mount-plugin
  def valid_dev(e)
    s = e[SPEC]
    t = e[TYPE]

    # has_valid_mount_device?
    valid = match_dev(s, '/dev/') #||
            # match_dev(s, "UUID="))

    if valid and File.symlink?(e[SPEC])
      e[SPEC] = File.expand_path([File.dirname(e[SPEC]),
                                  File.readlink(e[SPEC])].join '/')
    end

    #if (include_NFSs)
    valid = valid |
      (match_dev(t, "fuse") and !match_dev(t, "fuse.gvfs-fuse-daemon")) |
      match_dev(t, "shfs") |
      match_dev(t, "nfs") |
      match_dev(t, "cifs") |
      match_dev(t, "smbfs")

    if valid && e[MOUNTPOINT][0] == ?/ &&
        !match_dev(e[MOUNTPOINT], '/sys/')
      return true
    else
      return false
    end
  end
end

GIGA_SIZE = 1073741824.0
MEGA_SIZE = 1048576.0
KILO_SIZE = 1024.0

# Return the file size with a readable style.
def readable(size, precision=1)
  case
  when size < KILO_SIZE then "%d B" % size
  when size < MEGA_SIZE then "%.#{precision}f K" % (size / KILO_SIZE)
  when size < GIGA_SIZE then "%.#{precision}f M" % (size / MEGA_SIZE)
  else "%.#{precision}f GB" % (size / GIGA_SIZE)
  end
end

def refresh(menu)
  menu.each { |c| menu.remove(c) }

  $fstab = FsTab.new
  $fstab.each do |e|
    begin
      s = FileSystem.stat(e[FsTab::MOUNTPOINT])
      free = s.block_size * s.blocks_avail
      size = s.block_size * s.blocks
    rescue
      free = size = 0
    end

    #print readable(free), " "
    #print e[FsTab::MOUNTPOINT], " "
    #print e[FsTab::TYPE], " "
    #print e[FsTab::SPEC], "\n"

    item = Gtk::MenuItem.new
    h = Gtk::HBox.new

    l = Gtk::Label.new("#{e[FsTab::MOUNTPOINT]} ➔ #{e[FsTab::SPEC]}")
    l.set_width_chars(32)
    l.set_alignment(0, 0.5)
    h.pack_start(l)

    l = Gtk::Label.new("[#{readable(size-free)}/#{readable(size)}] #{readable(free)} free ")
    l.set_width_chars(28)
    l.set_alignment(1, 0.5)
    h.pack_start(l)

    p = Gtk::ProgressBar.new
    percent = size>0? 1.0-(free.to_f/size) : 0
    p.fraction = percent
    p.text = sprintf("%.1f %", percent*100)
    h.pack_start(p)

    item.add(h)
    menu.append(item)
    item.signal_connect('activate', &$activate)
    e << item
  end

  menu.show_all
end

class OSD_Mount
  def initialize
    @menu = Gtk::Menu.new
    # @m = Mutex.new
    # @menu.signal_connect('deactivate'){ @m.unlock }
  end

  def show(args)
    refresh(@menu)
    sleep(0.1)
    # this crashes without sleep, otherwise popup just don't work
    # @m.lock
    @menu.popup(nil, nil, 0, 0)
  end
end

class VolumeTray #< Gtk::StatusIcon
  def initialize
    @icons = VOLUME_ICONS
    @mute = VOLUME_MUTE
    volume = `amixer get PCM | sed -ne '/Front Left/s/.*\\[\\(.*\\)%\\].*/\\1/p'`
    volume = volume.strip.to_i
    @i = Gtk::StatusIcon.new
    @i.signal_connect('activate') {
      #Thread.new {
        if !system('ps aux|grep xfce4-mixer|grep -qv grep') then
          system('xfce4-mixer &')
        end
      #}
    }
    update(volume)
  end

  def update(v)
    case v
    when 0
      icon = @mute
    when 100
      icon = @icons.last
    else
      icon = @icons[ (v / (100/@icons.length.to_f)).to_i ]
    end
    @i.file = icon
    @i.tooltip = "PCM #{v}%"
  end
end

class OSD < Gtk::Window
  attr_accessor :w, :h
  attr_reader :stamp

  def initialize(icon_name)
    super(Gtk::Window::POPUP)
    type_hint = Gdk::Window::TYPE_HINT_NOTIFICATION
    decorated = true
    skip_taskbar_hint = true
    keep_above = true
    #window_position = Gtk::Window::POS_CENTER_ALWAYS

    @volume = Gtk::ProgressBar.new
    @frame = Gtk::Frame.new
    @frame.shadow_type = Gtk::SHADOW_OUT
    @vbox = Gtk::VBox.new
    @vbox.border_width = 12
    @vbox.spacing = 6

    @frame.add(@vbox)
    add(@frame)

    #require 'gnome2'
    #icon = Gtk::Image.new(Gnome::Stock::VOLUME, 6)
    @icon = Gtk::Image.new(icon_name)
    @icon.icon_size = 6

    @vbox.pack_start(@icon)
    @vbox.pack_start(@volume)

    show_all
    hide_all
    @w, @h = size
    @stamp = Time.at(0)
    @visible = false
  end

  def center
    screen, x, y, mask = Gdk::Display.default.pointer
    #geom = screen.monitor_geometry(0)
    monitor = screen.get_monitor(x, y)
    geom = screen.monitor_geometry(monitor)
    move(geom.x + ((geom.width-@w) / 2),
         geom.y + (geom.height / 2) + (geom.height / 2 - @h) / 2)
  end

  def show(args)
    @volume.fraction = args[0].to_f / 100
    @stamp = Time.now + 2

    center

    if !@visible
      show_all
      @visible = true
      GLib::Timeout.add_seconds(2) {
        if (@stamp <= Time.now)
          hide_all
          @visible = false
          while Gtk.events_pending?
            Gtk.main_iteration
          end
          false
        else
          true
        end
      }
    end
  end
end

class OSD_Volume < OSD
  def show(args)
    frac = args[0].to_f / 100

    if frac == 0 and @volume.fraction != 0
      @icon.file = VOLUME_MUTE
    elsif frac != 0 and @volume.fraction == 0
      @icon.file = VOLUME_OSD_ICON
    end

    @volume.fraction = frac
    @stamp = Time.now + 2

    center

    if !@visible
      show_all
      @visible = true
      GLib::Timeout.add_seconds(2) {
        if (@stamp <= Time.now)
          hide_all
          @visible = false
          while Gtk.events_pending?
            Gtk.main_iteration
          end
          false
        else
          true
        end
      }
    end
  end
end


## RUN

$osd = {
  'volume'     => [ OSD_Volume.new(VOLUME_OSD_ICON), VolumeTray.new ],
  'brightness' => [ OSD.new(BRIGHTNESS_OSD_ICON), nil],
  'mount'      => [ OSD_Mount.new ],
  'wm'         => [ Manager.new ]
}

# ugly
SOCKET_PATH = '/tmp/osd.socket'
if File.exists?(SOCKET_PATH)
  File.delete(SOCKET_PATH)
end

gs = UNIXServer.new(SOCKET_PATH)
#gs = TCPserver.open(12321)

require 'glib2'
gio = GLib::IOChannel.new(gs.fileno)
gio.add_watch(GLib::IOChannel::IN) do |chan, cond|
  begin
    s = gs.accept_nonblock
    s.gets
    s.close

    command, *args = $_.split
    osd, tray = $osd[command]

    if osd
      tray.update(args[0].to_i) if tray
      osd.show(args)
    end

  rescue Errno::EAGAIN, Errno::ECONNABORTED, Errno::EPROTO, Errno::EINTR
    retry
  rescue NoMethodError
  end

  true
end

Gtk.main
